import os
import re
import socket
import subprocess
import sys
from codecs import encode, decode

from pythonwhois.caching.whois_server_cache import server_cache
from pythonwhois.ratelimit.cool_down import CoolDown
from pythonwhois.response.raw_whois_response import RawWhoisResponse

incomplete_result_message = "THE_WHOIS_ORACLE_INCOMPLETE_RESULT"

cool_down_tracker = CoolDown()

# Sometimes IANA simply won't give us the right root WHOIS server
exceptions = {
    ".ac.uk": "whois.ja.net",
    ".ps": "whois.pnina.ps",
    ".buzz": "whois.nic.buzz",
    ".moe": "whois.nic.moe",
    ".arpa": "whois.iana.org",
    ".bid": "whois.nic.bid",
    ".int": "whois.iana.org",
    ".kred": "whois.nic.kred",
    ".nagoya": "whois.gmoregistry.net",
    ".nyc": "whois.nic.nyc",
    ".okinawa": "whois.gmoregistry.net",
    ".qpon": "whois.nic.qpon",
    ".sohu": "whois.gtld.knet.cn",
    ".tokyo": "whois.nic.tokyo",
    ".trade": "whois.nic.trade",
    ".webcam": "whois.nic.webcam",
    ".xn--rhqv96g": "whois.nic.xn--rhqv96g",
    # The following is a bit hacky, but IANA won't return the right answer for example.com because it's a direct registration.
    "example.com": "whois.verisign-grs.com"
}


def get_whois_raw(domain, server="", previous=None, rfc3490=True, never_cut=False, with_server_list=False,
                  server_list=None):
    previous = previous or []
    server_list = server_list or []

    if rfc3490:
        if sys.version_info < (3, 0):
            domain = encode(domain if type(domain) is unicode else decode(domain, "utf8"), "idna")
        else:
            domain = encode(domain, "idna").decode("ascii")

    target_server = get_target_server(domain, previous, server)
    query = prepare_query(target_server, domain)
    whois_response = query_server(target_server, query)
    response = whois_response.response

    if never_cut:
        # If the caller has requested to 'never cut' responses, he will get the original response from the server (this is
        # useful for callers that are only interested in the raw data). Otherwise, if the target is verisign-grs, we will
        # select the data relevant to the requested domain, and discard the rest, so that in a multiple-option response the
        # parsing code will only touch the information relevant to the requested domain. The side-effect of this is that
        # when `never_cut` is set to False, any verisign-grs responses in the raw data will be missing header, footer, and
        # alternative domain options (this is handled a few lines below, after the verisign-grs processing).
        new_list = [response] + previous
    if target_server == "whois.verisign-grs.com":
        # VeriSign is a little... special. As it may return multiple full records and there's no way to do an exact query,
        # we need to actually find the correct record in the list.
        for record in response.split("\n\n"):
            if re.search("Domain Name: %s\n" % domain.upper(), record):
                response = record
                break
    if not never_cut:
        new_list = [response] + previous

    if whois_response.server_is_dead:
        # That's probably as far as we can go, the road ends here
        return build_return_value(with_server_list, new_list, server_list)
    elif whois_response.request_failure:
        # Mark this result as incomplete, so we can try again later but still use the data if we have any
        new_list = [incomplete_result_message] + previous
        cool_down_tracker.warn_limit_exceeded(target_server)
        return build_return_value(with_server_list, new_list, server_list)
    elif whois_response.still_in_cool_down:
        new_list = [incomplete_result_message] + previous
        return build_return_value(with_server_list, new_list, server_list)

    server_list.append(target_server)

    # Ignore redirects from registries who publish the registrar data themselves
    if target_server not in ('whois.nic.xyz',):
        for line in [x.strip() for x in response.splitlines()]:
            match = re.match("(refer|whois server|referral url|whois server|registrar whois):\s*([^\s]+\.[^\s]+)", line,
                             re.IGNORECASE)
            if match is not None:
                referal_server = match.group(2)
                if referal_server != server and "://" not in referal_server \
                        and "www." not in referal_server and server_is_alive(referal_server):
                    # We want to ignore anything non-WHOIS (eg. HTTP) for now, and servers that are not reachable
                    # Referal to another WHOIS server...
                    return get_whois_raw(domain, referal_server, new_list, server_list=server_list,
                                         with_server_list=with_server_list)

    return build_return_value(with_server_list, new_list, server_list)


def build_return_value(with_server_list, responses, server_list):
    """
    Create a return value
    :param with_server_list: Whether the server list should be returned as well
    :param responses: The list of responses
    :param server_list: The server list
    :return: A list of responses without the empty ones, plus possibly a server list
    """
    non_empty_responses = filter((lambda text: text), responses)

    if with_server_list:
        return non_empty_responses, server_list
    else:
        return non_empty_responses


def query_server(whois_server, query):
    """
    Send out the query, if the server is available. if the server is still in cool down,
    return a RawWhoisResponse instance describing the failure
    :param whois_server: The WHOIS server to query
    :param query: The query to send
    :return: A RawWhoisResponse containing either the response or the reason of failure
    """
    if whois_server and cool_down_tracker.try_to_use_server(whois_server):
        return whois_request(query, whois_server)
    else:
        return RawWhoisResponse(still_in_cool_down=True)


def prepare_query(whois_server, domain):
    """
    Some WHOIS servers have a different way of querying.
    This methods returns an appropriate query for the WHOIS server
    :param domain: The domain to query
    :return: The fitting query
    """
    if whois_server == "whois.jprs.jp":
        return "%s/e" % domain  # Suppress Japanese output
    elif domain.endswith(".de") and (whois_server == "whois.denic.de" or whois_server == "de.whois-servers.net"):
        return "-T dn,ace %s" % domain  # regional specific stuff
    elif whois_server == "whois.verisign-grs.com":
        return "=%s" % domain  # Avoid partial matches
    else:
        return domain


def get_target_server(domain, previous_results, given_server):
    """
    Get the target server based on the current situation.
    :param domain: The domain to get the server for
    :param previous_results: The previously acquired results, as a result of referrals
    :param given_server:
    :return: The server to use
    """
    if len(previous_results) == 0 and given_server == "":
        # Root query
        for exception, exc_serv in exceptions.items():
            if domain.endswith(exception):
                target_server = exc_serv
                return target_server

        target_server = get_non_exception_server(domain)
        return target_server
    else:
        return given_server


def get_non_exception_server(domain):
    """
    Get a server that does not belong to the list of exceptions,
    either by asking IANA or by looking in the cache
    :param domain: The domain to get the WHOIS server for
    :return: The WHOIS server to use
    """
    tld = get_tld(domain)
    cached_server = server_cache.get_server(tld)
    if cached_server is not None:
        target_server = cached_server
    else:
        target_server = get_root_server(domain)
        server_cache.put_server(tld, target_server)

    return target_server


def server_is_alive(server):
    response = subprocess.call(["ping", "-c 1", "-w2", server], stdout=open(os.devnull, "w"),
                               stderr=subprocess.STDOUT)
    return response == 0


def get_tld(domain):
    return domain.split(".")[-1]


def get_root_server(domain):
    """
    Find the WHOIS server for a given domain
    :param domain: The domain to find a WHOIS server for
    :return: The WHOIS server, or an empty string if no server is found
    """
    data = whois_request(domain, "whois.iana.org").response or ""
    for line in [x.strip() for x in data.splitlines()]:
        match = re.match("refer:\s*([^\s]+)", line)
        if match is None:
            continue
        return match.group(1)
    return ""


def whois_request(domain, server, port=43, timeout=10):
    """
    Request WHOIS information.
    :param domain: The domain to request WHOIS information for
    :param server: The WHOIS server to use
    :param port: The port to use, 43 by default
    :param timeout: The length of the time out, 10 seconds by default
    :return: A WHOIS response containing either the result, or containing information about the failure
    """
    try:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(timeout)
        sock.connect((server, port))
        sock.send(("%s\r\n" % domain).encode("utf-8"))
        buff = b""
        while True:
            data = sock.recv(1024)
            if len(data) == 0:
                break
            buff += data
        return RawWhoisResponse(buff.decode("utf-8", "replace"))
    except Exception:
        server_is_dead = not server_is_alive(server)
        return RawWhoisResponse(request_failure=True, server_is_dead=server_is_dead)



import ConfigParser

from pythonwhois.ratelimit.cool_down_tracker import CoolDownTracker


def get_float_from_config(config, section, key, default=None):
    """
    Get a value from the config if it exists, otherwise return the default value
    :param config: The configuration to get the value from
    :param section: The section to get the value from
    :param key: The key that may or may not exist
    :param default: The default value to return, which is None by default
    :return: The value if it exists, else default
    """
    if config.has_option(section, key):
        return config.getfloat(section, key)
    else:
        return default


class CoolDownConfig:
    """
    Read and handle the contents of a configuration file for the cool down process.
    """

    def __init__(self, path_to_file, default_cool_down):
        """
        Read the configuration file.
        If the configuration contains a general section, this will be consumed and removed from the config instance
        (not the file). This is done to keep all the configuration in one file, but to be able to easily loop
        over all the WHOIS server sections.
        :param path_to_file: The path to the configuration file
        :param default_cool_down: The default value for the cool down length, in case it is not defined in the config
        """
        self.config = ConfigParser.ConfigParser()
        self.config.read(path_to_file)

        self.cool_down_length = default_cool_down
        self.config = self.consume_defaults_from_config(self.config)

    def consume_defaults_from_config(self, config):
        """
        Gets the general settings from the config. Then removes them
        and returns the modified config.
        :param config: The config to obtain the default values from
        :return: The modified config, without the 'general' section
        """
        if config.has_section("general"):
            self.cool_down_length = get_float_from_config(config, "general", "default_cool_down_length")
            config.remove_section("general")
        return config

    def get_sections(self):
        """
        Return a list of sections
        :return: A list of sections
        """
        return self.config.sections()

    def get_cool_down_tracker_for_server(self, whois_server):
        """
        Create a new CoolDownTracker instance based on the contents of the configuration file.
        If the configuration file does not have settings for this WHOIS server, a default CoolDownTracker with
        the cool down length is returned.
        :param whois_server: The WHOIS server to build a CoolDownTracker for
        :return: A CoolDownTracker instance based on the settings
        """
        if self.config.has_section(whois_server):
            cool_down_length = get_float_from_config(self.config, whois_server, "cool_down_length",
                                                     self.cool_down_length)
            max_requests_minute = get_float_from_config(self.config, whois_server, "max_requests_minute")
            max_requests_hour = get_float_from_config(self.config, whois_server, "max_requests_hour")
            max_requests_day = get_float_from_config(self.config, whois_server, "max_requests_day")
            return CoolDownTracker(cool_down_length,
                                   max_requests_minute,
                                   max_requests_hour,
                                   max_requests_day)

        return CoolDownTracker(self.cool_down_length)



import datetime

from pythonwhois.ratelimit.cool_down_config import CoolDownConfig
from pythonwhois.ratelimit.cool_down_tracker import CoolDownTracker


class CoolDown:
    """
    Handle the cool down period for asking a WHOIS server again
    """

    def __init__(self):
        """
        Creates a dictionary for storing cool downs.
        """
        self.servers_on_cool_down = {}
        self.default_cool_down_seconds = 2.0
        self.last_request_time = datetime.datetime.now()

    def can_use_server(self, whois_server):
        """
        Check whether a server can be used again
        :param whois_server: The WHOIS server to check
        :return: True if the server can be used, False if not
        """
        cool_down = self.servers_on_cool_down.get(whois_server)
        return cool_down is None or cool_down.current_cool_down <= 0

    def try_to_use_server(self, whois_server):
        """
        Try to use a WHOIS server. On True, it was a success and the cool down has been reset.
        On False, the server was not available yet
        :param whois_server: The WHOIS server that is going to be used
        :return True if the server was successfully marked as used and the cool down has been reset,
        False if the server was not yet available
        """
        self.decrement_cool_downs()
        if not self.can_use_server(whois_server):
            return False

        if whois_server not in self.servers_on_cool_down:
            self.servers_on_cool_down[whois_server] = CoolDownTracker(self.default_cool_down_seconds)
        self.servers_on_cool_down[whois_server].use_whois_server()
        return True

    def decrement_cool_downs(self):
        """
        Decrement all the cool downs with cool_down_time
        """
        time_diff = self.get_time_difference()
        for server, cool_down_tracker in self.servers_on_cool_down.iteritems():
            cool_down_tracker.decrement_cool_down(time_diff)

    def warn_limit_exceeded(self, whois_server):
        """
        Warn the CoolDown instance of an exceeded limit for a WHOIS server.
        The CoolDown instance will then make sure that the cool down for the WHOIS server
        will be longer next time
        :param whois_server: The WHOIS server the limit has been exceeded for
        """
        self.servers_on_cool_down[whois_server].double_cool_down()

    def get_time_difference(self):
        """
        Get the difference in time between te last time this was called
        and now.
        :return: The difference in seconds
        """
        now = datetime.datetime.now()
        diff = now - self.last_request_time
        self.last_request_time = now
        return diff.total_seconds()

    def set_cool_down_config(self, path_to_file):
        """
        Tell the CoolDown instance of a configuration file, describing specific settings
        for certain WHOIS servers. This configuration will then be read and inserted into
        the cool down dictionary.
        :param path_to_file: The path to the configuration file
        """
        cool_down_config = CoolDownConfig(path_to_file, self.default_cool_down_seconds)
        for whois_server in cool_down_config.get_sections():
            self.servers_on_cool_down[whois_server] = cool_down_config.get_cool_down_tracker_for_server(whois_server)



class RawWhoisResponse:
    """
    Holder class for WHOIS responses. Is capable of marking the retrieval as a failure.
    """

    def __init__(self, response="", request_failure=False, still_in_cool_down=False, server_is_dead=False):
        """
        Hold the WHOIS response
        :param response: The received response, if any
        :param request_failure: If the request was a failure
        :param still_in_cool_down: Whether the server was unavailable due to a cool down or not
        """
        self.response = response
        self.request_failure = request_failure
        self.still_in_cool_down = still_in_cool_down
        self.server_is_dead = server_is_dead

        if len(response) > 0:
            self.request_failure = self.check_for_exceeded_limit()

    def check_for_exceeded_limit(self):
        """
        Check whether the limit has been exceeded. This is done by
        looking at the size of the response. If it has less than 4 lines,
        it is probably not a useful response and most likely a message about spamming
        the WHOIS server
        :return: True if the message is really short, false if not
        """
        return self.response is not None and len(self.response.splitlines()) < 4



from setuptools import setup

setup(name='whois-oracle',
      version='1.1.1',
      description='Module for retrieving and parsing the WHOIS data for a domain. Supports most domains. No dependencies.',
      keywords='whois cool down',
      author='Sander ten Hoor, original by Sven Slootweg',
      url='https://github.com/MasterFenrir/whois-oracle',
      packages=['pythonwhois', 'pythonwhois.caching', 'pythonwhois.ratelimit', 'pythonwhois.response'],
      package_data={"pythonwhois": ["*.dat"]},
      install_requires=['argparse'],
      provides=['pythonwhois'],
      scripts=["whois-oracle"],
      license="WTFPL"
      )
